<?php

/**
 * @file
 * Provides a service to repair the dimensions of image field values.
 */

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Sql\SqlContentEntityStorage;
use Drupal\field\FieldStorageConfigInterface;
use Drupal\image\Plugin\Field\FieldType\ImageItem;
use Drupal\image_field_repair\ImageSizer;

/**
 * Callback that processes a partial batch run to repair image fields.
 *
 * Note: by only passing in the sandbox, this function can serve as a
 * hook_update_N() hook. This module will contain a wrapper around this function
 * to serve in form batch processing when this code is used in a stand alone
 * module.
 *
 * @param array $sandbox
 *   Array keeping track of the progress and other information that needs to be
 *   passed between subsequent calls to this batch operation callback function.
 *
 * @return \Drupal\Core\StringTranslation\TranslatableMarkup|\Drupal\Core\StringTranslation\TranslatableMarkup[]|null
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 */
function image_field_repair_process_batch(&$sandbox) {
  $batch_size = 100;
  $result = NULL;

  $entityTypeManager = \Drupal::entityTypeManager();
  $database = \Drupal::database();

  if (!isset($sandbox['field_ids'])) {
    // First run: initialize fields to repair.
    $sandbox['field_ids'] = image_field_repair_get_image_field_ids($entityTypeManager);
    $sandbox['table_infos'] = image_field_repair_get_all_table_infos($sandbox['field_ids'], $entityTypeManager, $database);
    $sandbox['current_table_index'] = 0;
    $sandbox['current_range_start'] = 0;
    $sandbox['total'] = array_sum(array_column($sandbox['table_infos'], 'record_count'));
    $sandbox['current_table_processed'] = 0;
    $sandbox['current_table_repaired'] = 0;
    $sandbox['processed'] = 0;
    $sandbox['repaired'] = 0;
    $sandbox['#finished'] = $sandbox['total'] !== 0 ? $sandbox['processed'] / $sandbox['total'] : 1;
  }
  else {
    // Process a limited set of records for the current field in the current
    // table starting at the current range start. If this results in fewer then
    // batch-size records processed, we advance to the next table but do not
    // start processing it in this pass.
    $table_info = $sandbox['table_infos'][$sandbox['current_table_index']];

    list($processed, $repaired) =  image_field_repair_process_table_range($table_info, $sandbox['current_range_start'], $batch_size, $database);

    // Update $sandbox.
    $sandbox['message'] = t('Processed records %from to %to of table %table for image field %field. %repaired of those had to be repaired.', [
      '%from' => $sandbox['processed'],
      '%to' => $sandbox['processed'] + $processed,
      '%table' => $sandbox['table_infos'][$sandbox['current_table_index']]['table_name'],
      '%field' => $sandbox['table_infos'][$sandbox['current_table_index']]['field_id'],
      '%repaired' => $repaired,
    ]);
    $sandbox['current_table_processed'] += $processed;
    $sandbox['current_table_repaired'] += $repaired;
    $sandbox['processed'] += $processed;
    $sandbox['repaired'] += $repaired;
    $sandbox['#finished'] = $sandbox['processed'] / $sandbox['total'];
    // If someone deleted an image during the update, current_table_processed
    // may never reach the stored record count, so we also check if the query
    // returned less than the requested number of records.
    if ($sandbox['current_table_processed'] >= $sandbox['table_infos'][$sandbox['current_table_index']]['record_count'] || $processed < $batch_size) {
      // End of this table: return statistics for this table and proceed to the
      // next table:
      if ($sandbox['current_table_repaired'] > 0) {
        $result = t('Repaired %repaired out of %total records in table "%table" for image field "%field".', [
          '%total' => $sandbox['current_table_processed'],
          '%table' => $sandbox['table_infos'][$sandbox['current_table_index']]['table_name'],
          '%field' => $sandbox['table_infos'][$sandbox['current_table_index']]['field_id'],
          '%repaired' => $sandbox['current_table_repaired'],
        ]);
      }

      $sandbox['current_table_index']++;
      $sandbox['current_range_start'] = 0;
      $sandbox['current_table_processed'] = 0;
      $sandbox['current_table_repaired'] = 0;

      // If someone deleted an image during the update, #finished will never
      // become 1, we so also indicate that we have finished when we have
      // processed all tables.
      if ($sandbox['current_table_index'] >= count($sandbox['table_infos'])) {
        $sandbox['#finished'] = 1;
      }
    }
    else {
      // Not yet at the end of this table: proceed to next batch size range.
      $sandbox['current_range_start'] += $processed;
      // Note: if the record count of this table is an exact multiple of the
      // batch size, we will eventually get a batch execution that does not
      // process any record.  However, as the current table index will be
      // incremented, this will not result in a never ending loop, so we do not
      // catch this situation.
    }

    // When we are finished we also return a summary message.
    if ($sandbox['#finished'] == 1) {
      $result = [
        $result,
        t('Total: Repaired %repaired out of %total checked records in %number_of_fields image fields.', [
          '%repaired' => $sandbox['repaired'],
          '%total' => $sandbox['processed'],
          '%number_of_fields' => count($sandbox['field_ids']),
        ]),
      ];
    }
  }

  return $result;
}

/**
 * Checks and repairs a range of records in a given image field table.
 *
 * @param array $table_info
 * @param int $rangeStart
 * @param int $rangeSize
 * @param \Drupal\Core\Database\Connection $database
 *
 * @return int[]
 *   2 integers indicating the number of records processed and the number of
 *   those that needed to be repaired.
 */
function image_field_repair_process_table_range(array $table_info, $rangeStart, $rangeSize, Connection $database) {
  $processed = 0;
  $repaired = 0;

  $table_name = $table_info['table_name'];
  $foreign_table_name = $table_info['join_field']['foreign_table_name'];
  // Query all columns of the image field (as we use almost all of them).
  $query = $database->select($table_name)
    ->fields($table_name);
  // Introduce a strict order by to ensure that the ranges do not miss any
  // record due to random/unknown ordering.
  foreach ($table_info['primary_key_fields'] as $primary_key_field) {
    $query->orderBy("$table_name.$primary_key_field");
  }
  // Join with the file_managed table as we need the image uri to get the actual
  // dimensions.
  $join_condition = sprintf('%s.%s = %s.%s', $foreign_table_name, $table_info['join_field']['foreign_field_name'],
    $table_name, $table_info['join_field']['join_field_name']);
  $query->join($foreign_table_name, NULL, $join_condition);
  $query->fields($foreign_table_name, ['uri']);
  // Add a range to stay well within the limits of batch processing.
  $query->range($rangeStart, $rangeSize);

  // Get and process the records.
  $results = $query->execute();
  while ($row = $results->fetchAssoc()) {
    // Get actual image dimensions and compare with stored image dimensions.
    // Do NOT use the image toolkit: we don't need it and we don't want to
    // depend on it or its current list of supported image types here.
    $image_size = ImageSizer::getSize($row['uri']);
    if (is_array($image_size)) {
      if ($row[$table_info['width_field']] != $image_size[0] || $row[$table_info['height_field']] != $image_size[1]) {
        // Dimensions are not equal: repair current record (identified by its
        // primary key values).
        $update_query = $database->update($table_name)
          ->fields(array(
            $table_info['width_field'] => $image_size[0],
            $table_info['height_field'] => $image_size[1],
          ));
        foreach ($table_info['primary_key_fields'] as $primary_key_field) {
          $update_query->condition($primary_key_field, $row[$primary_key_field]);
        }
        $repaired += $update_query->execute();
//        $repaired++;
      }
    }
    $processed++;
  }
  return [$processed, $repaired];
}

/**
 * Returns a list of image fields ids.
 *
 * We return all image fields, not only those that use a multi-upload widget
 * (required to be touched by issue [#2644468]) as that widget may have been
 * disabled due to this very issue.
 *
 * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
 *   The entity type manager.
 *
 * @return string[]
 *   A list of image fields ids.
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 */
function image_field_repair_get_image_field_ids(EntityTypeManagerInterface $entityTypeManager) {
  /** @var \Drupal\field\FieldStorageConfigStorage $storage */
  $storage = $entityTypeManager->getStorage('field_storage_config');
  $fieldIds = $storage->getQuery()
    ->condition('module', 'image')
    ->execute();
  return array_values($fieldIds);
}

/**
 * Returns info about the data and revision tables for all given image fields.
 *
 * @param string[] $fieldIds
 *   Array of ids of image fields.
 * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
 *   The entity type manager.
 * @param \Drupal\Core\Database\Connection $database
 *   The current database.
 *
 * @return array[]
 *   Table info about data and revision tables for all given image fields.
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 */
function image_field_repair_get_all_table_infos(array $fieldIds, EntityTypeManagerInterface $entityTypeManager, Connection $database) {
  $tables = [];
  $storage = $entityTypeManager->getStorage('field_storage_config');
  /** @var \Drupal\field\FieldStorageConfigInterface[] $fields */
  $fields = $storage->loadMultiple($fieldIds);
  foreach ($fields as $field) {
    $tables = array_merge($tables, image_field_repair_get_field_table_infos($field, $entityTypeManager, $database));
  }

  // Remove empty entries (for fields that do not use Sql storage or for non
  // revisioned fields).
  return array_values(array_filter($tables));
}

/**
 * Returns info about the data and revision tables for 1 image field.
 *
 * @param \Drupal\field\FieldStorageConfigInterface $field
 *   An image field.
 * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
 *   The entity type manager.
 * @param \Drupal\Core\Database\Connection $database
 *   The current database.
 *
 * @return array[]
 *   Info about the data and revision tables for the given image field.
 *   Will contain empty entries if the field is not stored in SqlStorage.
 *   Will contain an empty entry as 2nd entry if the field is not revisioned.
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 */
function image_field_repair_get_field_table_infos(FieldStorageConfigInterface $field, EntityTypeManagerInterface $entityTypeManager, Connection $database) {
  return [
    image_field_repair_get_table_info($field, TRUE, $entityTypeManager, $database),
    image_field_repair_get_table_info($field, FALSE, $entityTypeManager, $database)
  ];
}

/** @noinspection PhpDocMissingThrowsInspection */
/**
 * Returns the table info for 1 field table.
 *
 * @todo: to support other storage types we could check if the storage uses SQL
 *   tables ($entityTypeManager->getStorage($target_entity_type_id) returns a
 *   SqlContentEntityStorage) and if so, just use the database (faster),
 *   and if not so, use entity/field queries (slower and hooks are called which
 *   we may not want during an update).
 *
 * @param \Drupal\field\FieldStorageConfigInterface $field
 *   The field to get the table name for.
 * @param bool $dataOrRevision
 *   TRUE if we want the data table, FALSE if we want the revision table.
 * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
 *   The entity type manager.
 * @param \Drupal\Core\Database\Connection $database
 *   The current database.
 *
 * @return array
 *   An array with information about a field table that is needed later on in
 *   the process.
 *   An empty array is returned if no table could be found, either because the
 *   field does not use sql storage or because the field is not revisioned.
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 */
function image_field_repair_get_table_info(FieldStorageConfigInterface $field, $dataOrRevision, EntityTypeManagerInterface $entityTypeManager, Connection $database) {
  $order_by_fields = [
    'entity_id',
    'deleted',
    'delta',
    'langcode',
  ];
  $revision_order_by_fields = [
    'entity_id',
    'revision_id',
    'deleted',
    'delta',
    'langcode',
  ];

  $table_info = [];

  $target_entity_type_id = $field->getTargetEntityTypeId();
  $target_storage_definition = $entityTypeManager->getStorage($target_entity_type_id);
  // Ignore the field if it is not stored in Sql storage.
  if ($target_storage_definition instanceof SqlContentEntityStorage) {
    $table_mapping = $target_storage_definition->getTableMapping();
    if ($dataOrRevision) {
      $table_name = $table_mapping->getDedicatedDataTableName($field);
    }
    else {
      // Only if both the entity type the field is attached to, and the field
      // are revisionable, a revision table will exist.
      /** @noinspection PhpUnhandledExceptionInspection */
      $target_entity_type = $entityTypeManager->getDefinition($target_entity_type_id);
      if ($target_entity_type->isRevisionable() && $field->isRevisionable()) {
        $table_name = $table_mapping->getDedicatedRevisionTableName($field);
      }
    }

    if (!empty($table_name)) {
      // Get the join field, but change the property name (the array key in the
      // 'columns' entry) to the actual column name.
      $foreign_key = ImageItem::schema($field)['foreign keys']['target_id'];
      reset($foreign_key['columns']);
      $join_field = [
        'foreign_table_name' => $foreign_key['table'],
        'foreign_field_name' => current($foreign_key['columns']),
        'join_field_name' => $table_mapping->getFieldColumnName($field, key($foreign_key['columns'])),
      ];
      // Get the actual column names of the width and height properties.
      $width_field = $table_mapping->getFieldColumnName($field, 'width');
      $height_field = $table_mapping->getFieldColumnName($field, 'height');

      // Get the record count of this table.
      $record_count = $database->select($table_name)
        ->countQuery()
        ->execute()
        ->fetchField();

      // Only return this table if it is not empty.
      if ($record_count > 0) {
        $table_info = [
          'field_id' => $field->id(),
          'table_name' => $table_name,
          'record_count' => $record_count,
          'primary_key_fields' => $dataOrRevision ? $order_by_fields : $revision_order_by_fields,
          'join_field' => $join_field,
          'width_field' => $width_field,
          'height_field' => $height_field,
        ];
      }
    }
  }

  return $table_info;
}

/**
 * Returns the number of records for the given image fields.
 *
 * @param string[] $tableInfos
 *   Array of table infos about the image field tables.
 * @param \Drupal\Core\Database\Connection $database
 *
 * @return int
 *   The total number of records in the data and revision tables for the image
 *   fields.
 */
function image_field_repair_get_total_record_count(array $tableInfos, Connection $database) {
  $count = 0;
  foreach ($tableInfos as $tableInfo) {
    $count += $database->select($tableInfo['table_name'])
      ->countQuery()
      ->execute()
      ->fetchField();
  }
  return $count;
}
